/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2006-2016 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  $Id: tUsart.c 1353 2020-01-03 07:54:31Z ertl-komori $
 */

/*
 * LPC5500用 簡易SIOドライバ
 */

#include <sil.h>
#include "LPC5500.h"
#include "tUsart_tecsgen.h"

/*
 * シリアルI/Oポートのオープン
 */
void eSIOPort_open(CELLIDX idx)
{
    CELLCB *p_cellcb = GET_CELLCB(idx);

    /* 100 MHz / (2 * (1 + 240/256)) ~ 25.8 MHz */
    sil_wrw_mem(LPC5500_SYSCON_PLL0CLKDIV, 2 - 1);

    const uint32_t base  = ATTR_baseAddress;
    const uint32_t index = ATTR_index;

    sil_wrw_mem(LPC5500_SYSCON_FLEXFRG0CTRL, LPC5500_SYSCON_FLEXFRGCTRL_DIV(0xFF) | LPC5500_SYSCON_FLEXFRGCTRL_MULT(240));
    sil_wrw_mem(LPC5500_SYSCON_FCCLKSEL(index), LPC5500_SYSCON_FCCLKSEL_SEL_PLL0DIV); /* Select PLL0_div as Flexconn clock */
    sil_wrw_mem(LPC5500_SYSCON_AHBCLKCTRLSET1, LPC5500_SYSCON_AHBCLKCTRL1_FC(index)); /* Enable Flexconn clock */
    sil_wrw_mem(LPC5500_FLEXCOMM_PSELID(ATTR_flexcommAddress), LPC5500_FLEXCOMM_PSELID_PERSEL_USART); /* Select USART */

    /* 25.8 MHz / (16 * 14) ~ 115207 bps */
    /* More accurate but slowing peripheral option below */
    /* 100 MHz / (16 * 43 * (1 + 67/256)) ~ 115199 bps */
    sil_wrw_mem(LPC5500_USART_BRG(base), 14 - 1);
    sil_wrw_mem(LPC5500_USART_FIFOCFG(base), LPC5500_USART_FIFOCFG_ENABLETX | LPC5500_USART_FIFOCFG_ENABLERX); /* Enable FIFO */
    sil_wrw_mem(LPC5500_USART_FIFOTRIG(base),
                LPC5500_USART_FIFOTRIG_RXLVLENA | LPC5500_USART_FIFOTRIG_TXLVLENA | LPC5500_USART_FIFOTRIG_TXLVL(15));
             /*  Let RX generate FIFO interrupt  |  Let TX generate FIFO interrupt  | TX FIFO generates interrupt if there is a space */
    sil_wrw_mem(LPC5500_USART_CFG(base), LPC5500_USART_CFG_DATALEN_8BIT | LPC5500_USART_CFG_ENABLE);
                                        /*        8bit data length        |         Enable usart       */
}

/*
 * シリアルI/Oポートのクローズ
 */
void eSIOPort_close(CELLIDX idx)
{
    /*
     *  USARTをディスエーブル
     */
    CELLCB *p_cellcb = GET_CELLCB(idx);
    sil_wrw_mem(LPC5500_USART_CFG(ATTR_baseAddress), 0);
}

/*
 * シリアルI/Oポートへの文字送信
 */
bool_t eSIOPort_putChar(CELLIDX idx, char c)
{
    CELLCB *p_cellcb = GET_CELLCB(idx);
    if (sil_rew_mem(LPC5500_USART_FIFOSTAT(ATTR_baseAddress)) & LPC5500_USART_FIFOSTAT_TXNOTFULL) {
        sil_wrw_mem(LPC5500_USART_FIFOWR(ATTR_baseAddress), c);
        return true;
    }
    return false;
}

/*
 * シリアルI/Oポートからの文字受信
 */
int_t eSIOPort_getChar(CELLIDX idx)
{
    CELLCB *p_cellcb = GET_CELLCB(idx);
    if (sil_rew_mem(LPC5500_USART_FIFOSTAT(ATTR_baseAddress)) & LPC5500_USART_FIFOSTAT_RXNOTEMPTY) {
        return sil_rew_mem(LPC5500_USART_FIFORD(ATTR_baseAddress));
    }
    return -1;
}

/*
 * シリアルI/Oポートからのコールバックの許可
 */
void eSIOPort_enableCBR(CELLIDX idx, uint_t cbrtn)
{
    CELLCB *p_cellcb = GET_CELLCB(idx);
    switch (cbrtn) {
    case SIOSendReady:
        sil_wrw_mem(LPC5500_USART_FIFOINTENSET(ATTR_baseAddress), LPC5500_USART_FIFOINTEN_TXLVL); /* Enable TX FIFO interrupt */
        break;
    case SIOReceiveReady:
        sil_wrw_mem(LPC5500_USART_FIFOINTENSET(ATTR_baseAddress), LPC5500_USART_FIFOINTEN_RXLVL); /* Enable RX FIFO interrupt */
        break;
    }
}

/*
 * シリアルI/Oポートからのコールバックの禁止
 */
void eSIOPort_disableCBR(CELLIDX idx, uint_t cbrtn)
{
    CELLCB *p_cellcb = GET_CELLCB(idx);
    switch (cbrtn) {
    case SIOSendReady:
        sil_wrw_mem(LPC5500_USART_FIFOINTENCLR(ATTR_baseAddress), LPC5500_USART_FIFOINTEN_TXLVL);
        break;
    case SIOReceiveReady:
        sil_wrw_mem(LPC5500_USART_FIFOINTENCLR(ATTR_baseAddress), LPC5500_USART_FIFOINTEN_RXLVL);
        break;
    }
}

/*
 * シリアルI/Oポートに対する割込み処理
 */
void eiISR_main(CELLIDX idx)
{
    CELLCB *p_cellcb = GET_CELLCB(idx);

    const uint32_t stat = sil_rew_mem(LPC5500_USART_FIFOINTSTAT(ATTR_baseAddress));
    if (stat & LPC5500_USART_FIFOINTSTAT_TXLVL) {
        /*
         * 送信可能コールバックルーチンを呼び出す．
         */
        ciSIOCBR_readySend();
    }
    if (stat & LPC5500_USART_FIFOINTSTAT_RXLVL) {
        /*
         * 受信通知コールバックルーチンを呼び出す．
         */
        ciSIOCBR_readyReceive();
    }
}
