/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2016 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  $Id: target_timer.c 1363 2020-01-18 02:16:21Z ertl-komori $
 */

/*
 * タイマドライバ（CTimer用）
 * CTimer0を使用する．
 */

#include "kernel_impl.h"
#include "time_event.h"
#include "target_timer.h"
#include <sil.h>

/*
 * タイマの起動処理
 */
void target_hrt_initialize(intptr_t exinf)
{
    sil_wrw_mem(LPC5500_SYSCON_CTIMERCLKSEL0, LPC5500_SYSCON_CTIMERCLKSEL_SEL_MAIN_CLOCK); /* Select main clock as CTimer0 clock */
    sil_wrw_mem(LPC5500_SYSCON_AHBCLKCTRLSET1, LPC5500_SYSCON_AHBCLKCTRL1_TIMER0); /* Enable CTimer0 clock */

    sil_wrw_mem(LPC5500_CTIMER_PR(LPC5500_CTIMER0_BASE), 100 - 1); /* 100 MHz / 100 = 1 MHz */
    sil_wrw_mem(LPC5500_CTIMER_TCR(LPC5500_CTIMER0_BASE), LPC5500_CTIMER_TCR_CEN); /* Enable counter */
    
}

/*
 * タイマの停止処理
 */
void target_hrt_terminate(intptr_t exinf)
{
    /* Reset and stop timer */
    sil_wrw_mem(LPC5500_CTIMER_TCR(LPC5500_CTIMER0_BASE), LPC5500_CTIMER_TCR_CRST);
}

/*
 *  タイマ割込みハンドラ
 */
void target_hrt_handler(void)
{
    /* Clear interrupt flag */
    sil_wrw_mem(LPC5500_CTIMER_IR(LPC5500_CTIMER0_BASE), LPC5500_CTIMER_IR_MR0INT);
    /* Disable interrupt on match 0 */
    sil_wrw_mem(LPC5500_CTIMER_MCR(LPC5500_CTIMER0_BASE), 0);
    signal_time();
}
